/*

  Program First_order_lin:
  solve a first-order linear differential equation

  dy/dt = -ky

  with a Forward, Backward Euler and second order
  Runge-Kutta schemes, respectively.

 */

#include <iostream>
#include <cmath>
#include <fstream>
using namespace std;

main()
{

  cout << "Solution of equation: dy/dt = -k*y" << endl;
  cout << "through Forward-Euler, Backward-Euler and RK2" << endl;

  // Input of the parameters: time-step, final time, print time
  // (the latter MUST be a multiple of the time-step!)
  
  cout << "Values of dt, Tend, Tprint: ";
  double dt, Tend, Tprint;
  cin >> dt >> Tend >> Tprint;

  // Computes the total number of steps and the number of steps
  // between two consecutive prints
  
  long Nend   = long( Tend / dt );
  long Nprint = long( Tprint / dt );

  // Read the parameters and the initial conditions
  
  cout << "Parameter k: ";
  double k;
  cin >> k;

  cout << "Initial condition y0: ";
  double y0;
  cin >> y0;

  // File name for the output of data
  
  cout << "Name of the output file: ";
  char nfile[256];
  cin >> nfile;

  // Choice of the numerical scheme
  
  cout << "Choose the scheme you want to use:" << endl;
  cout << "  1 - Forward Euler" << endl;
  cout << "  2 - Backward Euler" << endl;
  cout << "  3 - 2nd order Runge-Kutta" << endl;
  int scheme;
  cin >> scheme;

  // Set the initial conditions, exact solution and error
  // The exact solution is given by: y(t) = y0 * exp(-k*t)

  double t_n, y_n, y_np1, t_star, y_star, y_ex, err;
  t_n  = 0.0;
  y_n  = y0;
  y_ex = y0;
  err = abs( y_n - y_ex );

  // Open the output file and write the initial condition

  ofstream fout;
  fout.open( nfile );
  fout.precision( 15 );
  fout << "#   t     y_num    y_ex    err" << endl;
  fout << t_n << "   " << y_n << "   " << y_ex << "   " << err << endl;

  // Loop in time

  for ( long n = 1; n <= Nend; n++ )
    {

      t_n = n * dt;

      // Compute y at next time step according to the scheme

      switch ( scheme )
	{
	case 1:    // Forward Euler
	  y_np1 = y_n * ( 1.0 - k * dt );
	  break;
	case 2:    // Backward Euler
	  y_np1 = y_n / ( 1.0 + k * dt );
	  break;
	case 3:
	  t_star = t_n + 0.5 * dt;
	  y_star = y_n + 0.5 * dt * ( -k * y_n );
	  y_np1  = y_n + dt * ( -k * y_star );
	  break;
	}

      y_n = y_np1;

      // Prints the solution each Tprint time intervals

      if ( ( n % Nprint ) == 0 )
	{
	  y_ex = y0 * exp( - k * t_n );
	  err = abs( y_n - y_ex );
	  fout << t_n << "   " << y_n << "   " << y_ex << "   " << err << endl;
	}

    }

  fout.close();

}
